import asyncio
import traceback
import uvicorn
from collections import namedtuple
from fastapi import FastAPI, Request
from fastapi.responses import RedirectResponse, FileResponse, Response, JSONResponse
from sovereign import (
    __version__,
    config,
    asgi_config,
    json_response_class,
    poller,
    template_context,
    logs,
)
from sovereign.error_info import ErrorInfo
from sovereign.views import crypto, discovery, healthchecks, admin, interface
from sovereign.middlewares import (
    RequestContextLogMiddleware,
    LoggingMiddleware,
)
from sovereign.utils.resources import get_package_file

Router = namedtuple("Router", "module tags prefix")

DEBUG = config.debug
SENTRY_DSN = config.sentry_dsn.get_secret_value()

try:
    import sentry_sdk
    from sentry_sdk.integrations.asgi import SentryAsgiMiddleware

    SENTRY_INSTALLED = True
except ImportError:  # pragma: no cover
    SENTRY_INSTALLED = False


def generic_error_response(e: Exception) -> JSONResponse:
    """
    Responds with a JSON object containing basic context
    about the exception passed in to this function.

    If the server is in debug mode, it will include a traceback in the response.

    The traceback is **always** emitted in logs.
    """
    tb = [line for line in traceback.format_exc().split("\n")]
    info = ErrorInfo.from_exception(e)
    logs.access_logger.queue_log_fields(
        ERROR=info.error,
        ERROR_DETAIL=info.detail,
        TRACEBACK=tb,
    )
    # Don't expose tracebacks in responses, but add it to the logs
    if DEBUG:
        info.traceback = tb
    return json_response_class(
        content=info.response, status_code=getattr(e, "status_code", 500)
    )


def init_app() -> FastAPI:
    application = FastAPI(
        title="Sovereign",
        version=__version__,
        debug=DEBUG,
        default_response_class=json_response_class,
    )

    routers = (
        Router(discovery.router, ["Configuration Discovery"], ""),
        Router(crypto.router, ["Cryptographic Utilities"], "/crypto"),
        Router(admin.router, ["Debugging Endpoints"], "/admin"),
        Router(interface.router, ["User Interface"], "/ui"),
        Router(healthchecks.router, ["Healthchecks"], ""),
    )
    for router in routers:
        application.include_router(
            router.module, tags=router.tags, prefix=router.prefix
        )

    application.add_middleware(RequestContextLogMiddleware)
    application.add_middleware(LoggingMiddleware)

    if SENTRY_INSTALLED and SENTRY_DSN:
        sentry_sdk.init(SENTRY_DSN)
        application.add_middleware(SentryAsgiMiddleware)
        logs.application_logger.logger.info("Sentry middleware enabled")

    @application.exception_handler(500)
    async def exception_handler(_: Request, exc: Exception) -> JSONResponse:
        """
        We cannot incur the execution of this function from unit tests
        because the starlette test client simply returns exceptions and does
        not run them through the exception handler.
        Ergo, this is a facade function for `generic_error_response`
        """
        return generic_error_response(exc)  # pragma: no cover

    @application.get("/")
    async def redirect_to_docs() -> Response:
        return RedirectResponse("/ui")

    @application.get("/static/{filename}", summary="Return a static asset")
    async def static(filename: str) -> Response:
        return FileResponse(get_package_file("sovereign", f"static/{filename}"))  # type: ignore[arg-type]

    @application.on_event("startup")
    async def keep_sources_uptodate() -> None:
        asyncio.create_task(poller.poll_forever())

    @application.on_event("startup")
    async def refresh_template_context() -> None:
        asyncio.create_task(template_context.start_refresh_context())

    return application


app = init_app()
logs.application_logger.logger.info(
    f"Sovereign started and listening on {asgi_config.host}:{asgi_config.port}"
)


if __name__ == "__main__":  # pragma: no cover
    uvicorn.run(app, host="0.0.0.0", port=8000, access_log=False)



from __future__ import division
from typing import List, Any, Dict, Generator


def _round_to_n(sequence: List[int], n: int = 100) -> List[int]:
    """Resolves issue when dividing by N results in a set of numbers that don't add up to N again

    E.g. 100 / 3 = 33
         33 * 3 = 99

    For the above example, this function returns [33, 33, 34]
    """
    if sum(sequence) == 0:
        return [0 for _ in sequence]
    if sum(sequence) != n:
        sequence[-1] = n - sum(sequence[:-1])
    return sequence


def _normalize_weights(sequence: List[int], n: int = 100) -> Generator[int, None, None]:
    total = sum(sequence)
    for item in sequence:
        try:
            yield int((item / total) * n)
        except ZeroDivisionError:
            yield 0


def fit_weights(
    clusters: List[Dict[str, Any]], total_weight: int = 100
) -> List[Dict[str, Any]]:
    weights = list(
        _normalize_weights([cluster["weight"] for cluster in clusters], n=total_weight)
    )
    for cluster, newly_assigned_weight in zip(
        clusters, _round_to_n(weights, n=total_weight)
    ):
        cluster["weight"] = newly_assigned_weight
    return clusters



from sovereign import template_context
from sovereign.modifiers.lib import Modifier
from sovereign.utils import eds, templates


class Test(Modifier):
    def match(self) -> bool:
        return True

    def apply(self) -> None:
        assert template_context
        assert eds
        assert templates
        self.instance["modifier_test_executed"] = True



import logging
from typing import Optional, Any, Callable, Dict
from functools import wraps
from sovereign.schemas import StatsdConfig

emitted: Dict[str, Any] = dict()


class StatsDProxy:
    def __init__(self, statsd_instance: Optional[Any] = None) -> None:
        self.statsd = statsd_instance
        self.emitted = emitted

    def __getattr__(self, item: str) -> Any:
        if self.statsd is not None:
            return getattr(self.statsd, item)
        try:
            return StatsdNoop
        except TypeError:
            return self.do_nothing

    def do_nothing(self, *args: Any, **kwargs: Any) -> None:
        k = args[0]
        emitted[k] = emitted.setdefault(k, 0) + 1


class StatsdNoop:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        k = args[0]
        emitted[k] = emitted.setdefault(k, 0) + 1

    def __enter__(self):  # type: ignore
        return self

    def __exit__(self, type: str, value: str, traceback: Any) -> None:
        pass

    def __call__(self, func: Callable[..., Any]):  # type: ignore
        @wraps(func)
        def wrapped(*args: Any, **kwargs: Any):  # type: ignore
            return func(*args, **kwargs)

        return wrapped


def configure_statsd(config: StatsdConfig) -> StatsDProxy:
    try:
        from datadog import DogStatsd

        class CustomStatsd(DogStatsd):  # type: ignore
            def _report(self, metric, metric_type, value, tags, sample_rate) -> None:  # type: ignore
                super()._report(metric, metric_type, value, tags, sample_rate)
                self.emitted: Dict[str, Any] = dict()
                self.emitted[metric] = self.emitted.setdefault(metric, 0) + 1

        module: Optional[CustomStatsd]
        module = CustomStatsd()
        if config.enabled:
            module.host = config.host
            module.port = config.port
            module.namespace = config.namespace
            module.use_ms = config.use_ms
            for tag, value in config.tags.items():
                module.constant_tags.extend([f"{tag}:{value}"])
        else:
            statsd_logger = logging.getLogger("datadog.dogstatsd")
            statsd_logger.disabled = True
            module = None
    except ImportError:
        if config.enabled:
            raise
        module = None

    return StatsDProxy(module)



import base64
import os

from cryptography.hazmat.primitives.ciphers.aead import AESGCM

from .base_cipher import CipherSuite


class AESGCMCipher(CipherSuite):
    AUTHENTICATE_DATA = "sovereign".encode()
    NONCE_LEN = 12

    def __str__(self) -> str:
        return "aesgcm"

    def __init__(self, secret_key: str):
        self.secret_key = base64.urlsafe_b64decode(secret_key)

    def encrypt(self, data: str) -> str:
        aesgcm = AESGCM(self.secret_key)
        nonce = os.urandom(self.NONCE_LEN)
        ct = aesgcm.encrypt(nonce, data.encode(), self.AUTHENTICATE_DATA)
        return base64.b64encode(nonce + ct).decode("utf-8")

    def decrypt(self, data: str) -> str:
        decoded_data = base64.b64decode(data)
        nonce, ct = (
            decoded_data[: self.NONCE_LEN],
            decoded_data[self.NONCE_LEN :],
        )
        aesgcm = AESGCM(self.secret_key)
        decrypted_output = aesgcm.decrypt(nonce, ct, self.AUTHENTICATE_DATA)
        return decrypted_output.decode("utf-8")

    @property
    def key_available(self) -> bool:
        return True

    @classmethod
    def generate_key(cls) -> bytes:
        # Generate 256 bit length key
        return base64.urlsafe_b64encode(os.urandom(32))



import json

from typing import Union, Optional, Any
from dataclasses import dataclass, field, asdict
from functools import singledispatchmethod

from sovereign import get_request_id

try:
    BOTO = True
    from botocore.exceptions import ClientError
except ImportError:
    BOTO = False


StrKeyDict = dict[str, Any]
Detail = Union[str, StrKeyDict]


@dataclass
class ErrorInfo:
    error: str
    detail: Detail
    request_id: str = field(default_factory=get_request_id)
    traceback: Optional[list[str]] = None

    @classmethod
    def _get_error(cls, exc: Exception, detail: Detail) -> "ErrorInfo":
        return cls(exc.__class__.__name__, detail)

    @singledispatchmethod
    @classmethod
    def from_exception(cls, exc: Exception) -> "ErrorInfo":
        return cls._get_error(exc, getattr(exc, "detail", "-"))

    if BOTO:

        @from_exception.register
        @classmethod
        def _(cls, exc: ClientError) -> Any:
            detail = {
                "message": str(exc),
                "operation": exc.operation_name,
                "response": exc.response,
            }

            return cls._get_error(exc, detail)

    @property
    def detail_str(self) -> str:
        if isinstance(self.detail, str):
            return self.detail
        return json.dumps(self.detail)

    @property
    def response(self) -> StrKeyDict:
        data = asdict(self)

        if data["traceback"] is None:
            del data["traceback"]

        return data



from functools import cache
import importlib.resources as res
from importlib.resources.abc import Traversable


@cache
def get_package(name: str) -> Traversable:
    return res.files(name)


def get_package_file(package_name: str, filename: str) -> Traversable:
    return get_package(package_name).joinpath(filename)


def get_package_file_bytes(package_name: str, filename: str) -> bytes:
    file = get_package_file(package_name, filename)
    return file.read_bytes()



from typing import List, Optional, Any, Dict
from socket import gethostbyname_ex
from socket import gaierror as dns_error
from starlette.exceptions import HTTPException
from sovereign import config, stats

REGIONS = config.legacy_fields.regions


def resolve(address: str) -> List[str]:
    try:
        with stats.timed("dns.resolve_ms", tags=[f"address:{address}"]):
            _, _, addresses = gethostbyname_ex(address)
    except dns_error:
        raise HTTPException(
            status_code=500, detail=f"Failed to resolve DNS hostname: {address}"
        )
    else:
        return addresses


def healthchecks_enabled(healthchecks: List[Dict[str, Any]]) -> bool:
    for healthcheck in healthchecks:
        if healthcheck.get("path") in ("no", False):
            return False
    return True


def upstream_requires_tls(cluster: Dict[str, Any]) -> bool:
    for host in cluster.get("hosts", []):
        if "443" in str(host.get("port")):
            return True
    return False


def list_regions() -> Optional[List[str]]:
    return REGIONS



import structlog
from structlog.exceptions import DropEvent
from structlog.stdlib import BoundLogger

from sovereign.logging.access_logger import AccessLogger
from sovereign.logging.application_logger import ApplicationLogger
from sovereign.logging.types import EventDict, ProcessedMessage
from sovereign.schemas import SovereignConfigv2


class LoggerBootstrapper:
    def __init__(self, config: SovereignConfigv2) -> None:
        self.show_debug: bool = config.debug

        structlog.configure(
            processors=[
                self.debug_logs_processor,
                structlog.processors.JSONRenderer(),
            ]
        )
        root_logger: BoundLogger = structlog.get_logger()
        self.logger = root_logger

        self.access_logger = AccessLogger(root_logger=root_logger, config=config)
        self.application_logger = ApplicationLogger(
            root_logger=root_logger, config=config
        )

    def debug_logs_processor(
        self, logger: BoundLogger, method_name: str, event_dict: EventDict
    ) -> ProcessedMessage:
        if method_name == "debug" and self.show_debug == False:
            raise DropEvent
        return event_dict



from typing import List
from fastapi import Response
from fastapi.routing import APIRouter
from fastapi.responses import PlainTextResponse
from sovereign import XDS_TEMPLATES, __version__, json_response_class
from sovereign.utils.mock import mock_discovery_request
from sovereign.views.discovery import perform_discovery


router = APIRouter()


@router.get("/healthcheck", summary="Healthcheck (Does the server respond to HTTP?)")
async def health_check() -> Response:
    return PlainTextResponse("OK")


@router.get(
    "/deepcheck",
    summary="Deepcheck (Can the server render all configured templates?)",
    response_class=json_response_class,
)
async def deep_check(response: Response) -> List[str]:
    response.status_code = 200
    ret = list()
    for template in list(XDS_TEMPLATES["default"].keys()):
        try:
            req = mock_discovery_request(service_cluster="*")
            await perform_discovery(req, "v3", resource_type=template, skip_auth=True)
        # pylint: disable=broad-except
        except Exception as e:
            ret.append(f"Failed {template}: {str(e)}")
        else:
            ret.append(f"Rendered {template} OK")
    return ret


@router.get("/version", summary="Display the current version of Sovereign")
async def version_check() -> Response:
    return PlainTextResponse(f"Sovereign {__version__}")



import base64
import os

from cryptography.fernet import Fernet

from .base_cipher import CipherSuite


class FernetCipher(CipherSuite):
    def __str__(self) -> str:
        return "fernet"

    def __init__(self, secret_key: str):
        self.fernet = Fernet(secret_key)

    def encrypt(self, data: str) -> str:
        return self.fernet.encrypt(data.encode()).decode("utf-8")

    def decrypt(self, data: str) -> str:
        return self.fernet.decrypt(data).decode("utf-8")

    @property
    def key_available(self) -> bool:
        return True

    @classmethod
    def generate_key(cls) -> bytes:
        # Generate 256 bit length key
        return base64.urlsafe_b64encode(os.urandom(32))



import asyncio
import traceback
from copy import deepcopy
from typing import (
    Any,
    Awaitable,
    Dict,
    Generator,
    Iterable,
    NamedTuple,
    NoReturn,
    Optional,
)

from fastapi import HTTPException
from structlog.stdlib import BoundLogger

from sovereign.config_loader import Loadable
from sovereign.schemas import DiscoveryRequest, EncryptionConfig, XdsTemplate
from sovereign.sources import SourcePoller
from sovereign.utils.crypto.crypto import CipherContainer
from sovereign.utils.crypto.suites import EncryptionType
from sovereign.utils.timer import poll_forever, poll_forever_cron


class LoadContextResponse(NamedTuple):
    context_name: str
    context: Dict[str, Any]
    success: bool = True


class TemplateContext:
    def __init__(
        self,
        refresh_rate: Optional[int],
        refresh_cron: Optional[str],
        refresh_num_retries: int,
        refresh_retry_interval_secs: int,
        configured_context: Dict[str, Loadable],
        poller: SourcePoller,
        encryption_suite: Optional[CipherContainer],
        logger: BoundLogger,
        stats: Any,
    ) -> None:
        self.poller = poller
        self.refresh_rate = refresh_rate
        self.refresh_cron = refresh_cron
        self.refresh_num_retries = refresh_num_retries
        self.refresh_retry_interval_secs = refresh_retry_interval_secs
        self.configured_context = configured_context
        self.crypto: CipherContainer | None = encryption_suite
        self.logger = logger
        self.stats = stats
        # initial load
        self.context: Dict[str, Any] = {}
        asyncio.run(self.load_context_variables())

    async def start_refresh_context(self) -> NoReturn:
        if self.refresh_cron is not None:
            await poll_forever_cron(self.refresh_cron, self.refresh_context)
        elif self.refresh_rate is not None:
            await poll_forever(self.refresh_rate, self.refresh_context)

        raise RuntimeError("Failed to start refresh_context, this should never happen")

    async def refresh_context(self) -> None:
        await self.load_context_variables()

    async def _load_context(
        self,
        context_name: str,
        context_config: Loadable | str,
        refresh_num_retries: int,
        refresh_retry_interval_secs: int,
    ) -> LoadContextResponse:
        retries_left = refresh_num_retries
        context_response = {}

        while True:
            try:
                if isinstance(context_config, Loadable):
                    context_response = context_config.load()
                elif isinstance(context_config, str):
                    context_response = Loadable.from_legacy_fmt(context_config).load()
                self.stats.increment(
                    "context.refresh.success",
                    tags=[f"context:{context_name}"],
                )
                return LoadContextResponse(context_name, context_response)
            # pylint: disable=broad-except
            except Exception as e:
                retries_left -= 1
                if retries_left < 0:
                    tb = [line for line in traceback.format_exc().split("\n")]
                    self.logger.error(str(e), traceback=tb)
                    self.stats.increment(
                        "context.refresh.error",
                        tags=[f"context:{context_name}"],
                    )
                    return LoadContextResponse(context_name, context_response, False)
                else:
                    await asyncio.sleep(refresh_retry_interval_secs)

    async def load_context_variables(self) -> None:
        context_coroutines: list[Awaitable[LoadContextResponse]] = []
        for context_name, context_config in self.configured_context.items():
            context_coroutines.append(
                self._load_context(
                    context_name,
                    context_config,
                    self.refresh_num_retries,
                    self.refresh_retry_interval_secs,
                )
            )

        context_results: list[LoadContextResponse] = await asyncio.gather(
            *context_coroutines
        )

        for context_result in context_results:
            if (
                context_result.success
                or context_result.context_name not in self.context
            ):
                self.context[context_result.context_name] = context_result.context

        if "crypto" not in self.context and self.crypto:
            self.context["crypto"] = self.crypto

    def build_new_context_from_instances(self, node_value: str) -> Dict[str, Any]:
        matches = self.poller.match_node(node_value=node_value)
        ret = dict()
        for key, value in self.context.items():
            try:
                ret[key] = deepcopy(value)
            except TypeError:
                ret[key] = value

        to_add = dict()
        for scope, instances in matches.scopes.items():
            if scope in ("default", None):
                to_add["instances"] = instances
            else:
                to_add[scope] = instances
        if to_add == {}:
            raise HTTPException(
                detail=(
                    "This node does not match any instances! ",
                    "If node matching is enabled, check that the node "
                    "match key aligns with the source match key. "
                    "If you don't know what any of this is, disable "
                    "node matching via the config",
                ),
                status_code=400,
            )
        ret.update(to_add)
        return ret

    def get_context(
        self, request: DiscoveryRequest, template: XdsTemplate
    ) -> Dict[str, Any]:
        ret = self.build_new_context_from_instances(
            node_value=self.poller.extract_node_key(request.node),
        )
        if request.hide_private_keys:
            ret["crypto"] = CipherContainer.from_encryption_configs(
                encryption_configs=[EncryptionConfig("", EncryptionType.DISABLED)],
                logger=self.logger,
            )
        if not template.is_python_source:
            keys_to_remove = self.unused_variables(list(ret), template.jinja_variables)
            for key in keys_to_remove:
                ret.pop(key, None)
        return ret

    @staticmethod
    def unused_variables(
        keys: Iterable[str], variables: Iterable[str]
    ) -> Generator[str, None, None]:
        for key in keys:
            if key not in variables:
                yield key

    def get(self, *args: Any, **kwargs: Any) -> Any:
        return self.context.get(*args, **kwargs)



from typing import Any

from .base_cipher import CipherSuite


class DisabledCipher(CipherSuite):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def __str__(self) -> str:
        return "disabled"

    def encrypt(self, _: str) -> str:
        return "Unavailable (No Secret Key)"

    def decrypt(self, _: str) -> str:
        return "Unavailable (No Secret Key)"

    @property
    def key_available(self) -> bool:
        return False

    @classmethod
    def generate_key(cls) -> bytes:
        return b"Unavailable (No key to generate)"



import multiprocessing
import warnings
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from os import getenv
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple, Type, Union

from croniter import CroniterBadCronError, croniter
from fastapi.responses import JSONResponse
from jinja2 import Template, meta
from pydantic import (
    BaseModel,
    BaseSettings,
    Field,
    SecretStr,
    root_validator,
    validator,
)

from sovereign.config_loader import Loadable, Protocol, Serialization, jinja_env
from sovereign.utils.crypto.suites import EncryptionType
from sovereign.utils.version_info import compute_hash

missing_arguments = {"missing", "positional", "arguments:"}

JsonResponseClass: Type[JSONResponse] = JSONResponse
# pylint: disable=unused-import
try:
    import orjson
    from fastapi.responses import ORJSONResponse

    JsonResponseClass = ORJSONResponse
except ImportError:
    try:
        import ujson
        from fastapi.responses import UJSONResponse

        JsonResponseClass = UJSONResponse
    except ImportError:
        pass


class CacheStrategy(str, Enum):
    context = "context"
    content = "content"


class SourceData(BaseModel):
    scopes: Dict[str, List[Dict[str, Any]]] = defaultdict(list)


class ConfiguredSource(BaseModel):
    type: str
    config: Dict[str, Any]
    scope: str = "default"  # backward compatibility


class StatsdConfig(BaseModel):
    host: str = "127.0.0.1"
    port: Union[int, str] = 8125
    tags: Dict[str, Union[Loadable, str]] = dict()
    namespace: str = "sovereign"
    enabled: bool = False
    use_ms: bool = True

    @validator("host", pre=True)
    def load_host(cls, v: str) -> Any:
        return Loadable.from_legacy_fmt(v).load()

    @validator("port", pre=True)
    def load_port(cls, v: Union[int, str]) -> Any:
        if isinstance(v, int):
            return v
        elif isinstance(v, str):
            return Loadable.from_legacy_fmt(v).load()
        else:
            raise ValueError(f"Received an invalid port: {v}")

    @validator("tags", pre=True)
    def load_tags(cls, v: Dict[str, Union[Loadable, str]]) -> Dict[str, Any]:
        ret = dict()
        for key, value in v.items():
            if isinstance(value, dict):
                ret[key] = Loadable(**value).load()
            elif isinstance(value, str):
                ret[key] = Loadable.from_legacy_fmt(value).load()
            else:
                raise ValueError(f"Received an invalid tag for statsd: {value}")
        return ret


class DiscoveryCacheConfig(BaseModel):
    enabled: bool = False
    host: str = "localhost"
    port: int = 6379
    secure: bool = False
    protocol: str = "redis://"
    password: SecretStr = SecretStr("")
    client_side: bool = True  # True = Try in-memory cache before hitting redis
    wait_for_connection_timeout: int = 5
    socket_connect_timeout: int = 5
    socket_timeout: int = 5
    max_connections: int = 100
    retry_on_timeout: bool = True  # Retry connections if they timeout.
    suppress: bool = False  # False = Don't suppress connection errors. True = suppress connection errors
    socket_keepalive: bool = True  # Try to keep connections to redis around.
    ttl: int = 60

    @root_validator
    def set_default_protocol(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        secure = values.get("secure")
        if secure:
            values["protocol"] = "rediss://"
        return values

    @root_validator
    def set_environmental_variables(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if host := getenv("SOVEREIGN_DISCOVERY_CACHE_REDIS_HOST"):
            values["host"] = host
        if port := getenv("SOVEREIGN_DISCOVERY_CACHE_REDIS_PORT"):
            values["port"] = int(port)
        if password := getenv("SOVEREIGN_DISCOVERY_CACHE_REDIS_PASSWORD"):
            values["password"] = SecretStr(password)
        return values


class XdsTemplate:
    def __init__(self, path: Union[str, Loadable]) -> None:
        if isinstance(path, str):
            self.loadable: Loadable = Loadable.from_legacy_fmt(path)
        elif isinstance(path, Loadable):
            self.loadable = path
        self.is_python_source = self.loadable.protocol == Protocol.python
        self.source = self.load_source()
        template_ast = jinja_env.parse(self.source)
        self.jinja_variables = meta.find_undeclared_variables(template_ast)

    def __call__(
        self, *args: Any, **kwargs: Any
    ) -> Optional[Union[Dict[str, Any], str]]:
        if not hasattr(self, "code"):
            self.code: Union[Template, ModuleType] = self.loadable.load()
        if isinstance(self.code, ModuleType):
            try:
                return {"resources": list(self.code.call(*args, **kwargs))}
            except TypeError as e:
                if not set(str(e).split()).issuperset(missing_arguments):
                    raise e
                message_start = str(e).find(":")
                missing_args = str(e)[message_start + 2 :]
                supplied_args = list(kwargs.keys())
                raise TypeError(
                    f"Tried to render a template using partial arguments. "
                    f"Missing args: {missing_args}. Supplied args: {args} "
                    f"Supplied keyword args: {supplied_args}"
                )
        else:
            return self.code.render(*args, **kwargs)

    def load_source(self) -> str:
        if self.loadable.serialization in (Serialization.jinja, Serialization.jinja2):
            # The Jinja2 template serializer does not properly set a name
            # for the loaded template.
            # The repr for the template prints out as the memory address
            # This makes it really hard to generate a consistent version_info string
            # in rendered configuration.
            # For this reason, we re-load the template as a string instead, and create a checksum.
            old_serialization = self.loadable.serialization
            self.loadable.serialization = Serialization("string")
            ret = self.loadable.load()
            self.loadable.serialization = old_serialization
            return str(ret)
        elif self.is_python_source:
            # If the template specified is a python source file,
            # we can simply read and return the source of it.
            old_protocol = self.loadable.protocol
            old_serialization = self.loadable.serialization
            self.loadable.protocol = Protocol("inline")
            self.loadable.serialization = Serialization("string")
            ret = self.loadable.load()
            self.loadable.protocol = old_protocol
            self.loadable.serialization = old_serialization
            return str(ret)
        ret = self.loadable.load()
        return str(ret)

    def __repr__(self) -> str:
        return f"XdsTemplate({self.loadable=}, {self.is_python_source=}, {self.source=}, {self.jinja_variables=})"


class ProcessedTemplate:
    def __init__(
        self,
        resources: List[Dict[str, Any]],
        version_info: Optional[str],
    ) -> None:
        self.resources = resources
        self.version_info = version_info
        self._rendered: Optional[bytes] = None

    @property
    def version(self) -> str:
        return self.version_info or compute_hash(self.resources)

    @property
    def rendered(self) -> bytes:
        if self._rendered is None:
            result = JsonResponseClass(content="").render(
                content={
                    "version_info": self.version,
                    "resources": self.resources,
                }
            )
            self._rendered = result
        return self._rendered

    def deserialize_resources(self) -> List[Dict[str, Any]]:
        return self.resources


class Locality(BaseModel):
    region: str = Field(None)
    zone: str = Field(None)
    sub_zone: str = Field(None)


class SemanticVersion(BaseModel):
    major_number: int = 0
    minor_number: int = 0
    patch: int = 0

    def __str__(self) -> str:
        return f"{self.major_number}.{self.minor_number}.{self.patch}"


class BuildVersion(BaseModel):
    version: SemanticVersion = SemanticVersion()
    metadata: Dict[str, Any] = {}


class Extension(BaseModel):
    name: Optional[str] = None
    category: Optional[str] = None
    version: Optional[BuildVersion] = None
    disabled: Optional[bool] = None


class Node(BaseModel):
    id: str = Field("-", title="Hostname")
    cluster: str = Field(
        ...,
        title="Envoy service-cluster",
        description="The ``--service-cluster`` configured by the Envoy client",
    )
    metadata: Dict[str, Any] = Field(default_factory=dict, title="Key:value metadata")
    locality: Locality = Field(Locality(), title="Locality")  # type: ignore
    build_version: str = Field(
        None,  # Optional in the v3 Envoy API
        title="Envoy build/release version string",
        description="Used to identify what version of Envoy the "
        "client is running, and what config to provide in response",
    )
    user_agent_name: str = "envoy"
    user_agent_version: str = ""
    user_agent_build_version: BuildVersion = BuildVersion()
    extensions: List[Extension] = []
    client_features: List[str] = []

    @property
    def common(self) -> Tuple[str, str, str, BuildVersion, Locality]:
        """
        Returns fields that are the same in adjacent proxies
        ie. proxies that are part of the same logical group
        """
        return (
            self.cluster,
            self.build_version,
            self.user_agent_version,
            self.user_agent_build_version,
            self.locality,
        )


class Resources(List[str]):
    """
    Acts like a regular list except it returns True
    for all membership tests when empty.
    """

    def __contains__(self, item: object) -> bool:
        if len(self) == 0:
            return True
        return item in list(self)


class Status(BaseModel):
    code: int
    message: str
    details: List[Any]


class DiscoveryRequest(BaseModel):
    node: Node = Field(..., title="Node information about the envoy proxy")
    version_info: str = Field(
        "0", title="The version of the envoy clients current configuration"
    )
    resource_names: Resources = Field(
        Resources(), title="List of requested resource names"
    )
    hide_private_keys: bool = False
    type_url: Optional[str] = Field(
        None, title="The corresponding type_url for the requested resource"
    )
    desired_controlplane: str = Field(
        None, title="The host header provided in the Discovery Request"
    )
    error_detail: Status = Field(
        None, title="Error details from the previous xDS request"
    )

    @property
    def envoy_version(self) -> str:
        try:
            version = str(self.node.user_agent_build_version.version)
            assert version != "0.0.0"
        except AssertionError:
            try:
                build_version = self.node.build_version
                _, version, *_ = build_version.split("/")
            except (AttributeError, ValueError):
                # TODO: log/metric this?
                return "default"
        return version

    @property
    def resources(self) -> Resources:
        return Resources(self.resource_names)

    @property
    def uid(self) -> str:
        return compute_hash(
            self.resources,
            self.node.common,
            self.desired_controlplane,
        )


class DiscoveryResponse(BaseModel):
    version_info: str = Field(
        ..., title="The version of the configuration in the response"
    )
    resources: List[Any] = Field(..., title="The requested configuration resources")


class SovereignAsgiConfig(BaseSettings):
    host: str = "0.0.0.0"
    port: int = 8080
    keepalive: int = 5
    workers: int = multiprocessing.cpu_count() * 2 + 1
    threads: int = 1
    reuse_port: bool = True
    preload_app: bool = True
    log_level: str = "warning"
    worker_class: str = "uvicorn.workers.UvicornWorker"
    worker_timeout: int = 30
    worker_tmp_dir: str = "/dev/shm"
    graceful_timeout: int = worker_timeout * 2

    class Config:
        fields = {
            "host": {"env": "SOVEREIGN_HOST"},
            "port": {"env": "SOVEREIGN_PORT"},
            "keepalive": {"env": "SOVEREIGN_KEEPALIVE"},
            "workers": {"env": "SOVEREIGN_WORKERS"},
            "threads": {"env": "SOVEREIGN_THREADS"},
            "preload_app": {"env": "SOVEREIGN_PRELOAD"},
            "worker_timeout": {"env": "SOVEREIGN_WORKER_TIMEOUT"},
        }

    def as_gunicorn_conf(self) -> Dict[str, Any]:
        return {
            "bind": ":".join(map(str, [self.host, self.port])),
            "keepalive": self.keepalive,
            "reuse_port": self.reuse_port,
            "preload_app": self.preload_app,
            "loglevel": self.log_level,
            "timeout": self.worker_timeout,
            "threads": self.threads,
            "workers": self.workers,
            "worker_class": self.worker_class,
            "worker_tmp_dir": self.worker_tmp_dir,
            "graceful_timeout": self.graceful_timeout,
        }


class SovereignConfig(BaseSettings):
    sources: List[ConfiguredSource]
    templates: Dict[str, Dict[str, Union[str, Loadable]]]
    template_context: Dict[str, Any] = {}
    eds_priority_matrix: Dict[str, Dict[str, str]] = {}
    modifiers: List[str] = []
    global_modifiers: List[str] = []
    regions: List[str] = []
    statsd: StatsdConfig = StatsdConfig()
    auth_enabled: bool = False
    auth_passwords: str = ""
    encryption_key: str = ""
    environment: str = "local"
    debug_enabled: bool = False
    sentry_dsn: str = ""
    node_match_key: str = "cluster"
    node_matching: bool = True
    source_match_key: str = "service_clusters"
    sources_refresh_rate: int = 30
    cache_strategy: str = "context"
    refresh_context: bool = False
    context_refresh_rate: Optional[int]
    context_refresh_cron: Optional[str]
    dns_hard_fail: bool = False
    enable_application_logs: bool = True
    enable_access_logs: bool = True
    log_fmt: Optional[str] = ""
    ignore_empty_log_fields: bool = False
    discovery_cache: DiscoveryCacheConfig = DiscoveryCacheConfig()

    class Config:
        fields = {
            "auth_enabled": {"env": "SOVEREIGN_AUTH_ENABLED"},
            "auth_passwords": {"env": "SOVEREIGN_AUTH_PASSWORDS"},
            "encryption_key": {"env": "SOVEREIGN_ENCRYPTION_KEY"},
            "environment": {"env": "SOVEREIGN_ENVIRONMENT"},
            "debug_enabled": {"env": "SOVEREIGN_DEBUG_ENABLED"},
            "sentry_dsn": {"env": "SOVEREIGN_SENTRY_DSN"},
            "node_match_key": {"env": "SOVEREIGN_NODE_MATCH_KEY"},
            "node_matching": {"env": "SOVEREIGN_NODE_MATCHING"},
            "source_match_key": {"env": "SOVEREIGN_SOURCE_MATCH_KEY"},
            "sources_refresh_rate": {"env": "SOVEREIGN_SOURCES_REFRESH_RATE"},
            "cache_strategy": {"env": "SOVEREIGN_CACHE_STRATEGY"},
            "refresh_context": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
            "context_refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
            "context_refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
            "dns_hard_fail": {"env": "SOVEREIGN_DNS_HARD_FAIL"},
            "enable_application_logs": {"env": "SOVEREIGN_ENABLE_APPLICATION_LOGS"},
            "enable_access_logs": {"env": "SOVEREIGN_ENABLE_ACCESS_LOGS"},
            "log_fmt": {"env": "SOVEREIGN_LOG_FORMAT"},
            "ignore_empty_fields": {"env": "SOVEREIGN_LOG_IGNORE_EMPTY"},
        }

    @property
    def passwords(self) -> List[str]:
        return self.auth_passwords.split(",") or []

    def xds_templates(self) -> Dict[str, Dict[str, XdsTemplate]]:
        ret: Dict[str, Dict[str, XdsTemplate]] = {
            "__any__": {}
        }  # Special key to hold templates from all versions
        for version, templates in self.templates.items():
            loaded_templates = {
                _type: XdsTemplate(path=path) for _type, path in templates.items()
            }
            ret[str(version)] = loaded_templates
            ret["__any__"].update(loaded_templates)
        return ret

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        kwargs = [f"{k}={v}" for k, v in self.show().items()]
        return f"SovereignConfig({kwargs})"

    def show(self) -> Dict[str, Any]:
        safe_items = dict()
        for key, value in self.__dict__.items():
            if key in ["auth_passwords", "encryption_key", "passwords", "sentry_dsn"]:
                value = "redacted"
            safe_items[key] = value
        return safe_items


class TemplateSpecification(BaseModel):
    type: str
    spec: Loadable


class NodeMatching(BaseSettings):
    enabled: bool = True
    source_key: str = "service_clusters"
    node_key: str = "cluster"

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_NODE_MATCHING_ENABLED"},
            "source_key": {"env": "SOVEREIGN_SOURCE_MATCH_KEY"},
            "node_key": {"env": "SOVEREIGN_NODE_MATCH_KEY"},
        }


@dataclass
class EncryptionConfig:
    encryption_key: str
    encryption_type: EncryptionType


class AuthConfiguration(BaseSettings):
    enabled: bool = False
    auth_passwords: SecretStr = SecretStr("")
    encryption_key: SecretStr = SecretStr("")

    @staticmethod
    def _create_encryption_config(encryption_key_setting: str) -> EncryptionConfig:
        encryption_key, _, encryption_type_raw = encryption_key_setting.partition(":")
        if encryption_type_raw:
            encryption_type = EncryptionType(encryption_type_raw)
        else:
            encryption_type = EncryptionType.FERNET
        return EncryptionConfig(encryption_key, encryption_type)

    @property
    def encryption_configs(self) -> tuple[EncryptionConfig, ...]:
        secret_values = self.encryption_key.get_secret_value().split()

        configs = tuple(
            self._create_encryption_config(encryption_key_setting)
            for encryption_key_setting in secret_values
        )
        return configs

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_AUTH_ENABLED"},
            "auth_passwords": {"env": "SOVEREIGN_AUTH_PASSWORDS"},
            "encryption_key": {"env": "SOVEREIGN_ENCRYPTION_KEY"},
        }


class ApplicationLogConfiguration(BaseSettings):
    enabled: bool = False
    log_fmt: Optional[str] = None
    # currently only support /dev/stdout as JSON

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_ENABLE_APPLICATION_LOGS"},
            "log_fmt": {"env": "SOVEREIGN_APPLICATION_LOG_FORMAT"},
        }


class AccessLogConfiguration(BaseSettings):
    enabled: bool = True
    log_fmt: Optional[str] = None
    ignore_empty_fields: bool = False

    class Config:
        fields = {
            "enabled": {"env": "SOVEREIGN_ENABLE_ACCESS_LOGS"},
            "log_fmt": {"env": "SOVEREIGN_LOG_FORMAT"},
            "ignore_empty_fields": {"env": "SOVEREIGN_LOG_IGNORE_EMPTY"},
        }


class LoggingConfiguration(BaseSettings):
    application_logs: ApplicationLogConfiguration = ApplicationLogConfiguration()
    access_logs: AccessLogConfiguration = AccessLogConfiguration()


class ContextConfiguration(BaseSettings):
    context: Dict[str, Loadable] = {}
    refresh: bool = False
    refresh_rate: Optional[int] = None
    refresh_cron: Optional[str] = None
    refresh_num_retries: int = 3
    refresh_retry_interval_secs: int = 10

    @staticmethod
    def context_from_legacy(context: Dict[str, str]) -> Dict[str, Loadable]:
        ret = dict()
        for key, value in context.items():
            ret[key] = Loadable.from_legacy_fmt(value)
        return ret

    @root_validator(pre=False)
    def validate_single_use_refresh_method(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        refresh_rate = values.get("refresh_rate")
        refresh_cron = values.get("refresh_cron")

        if (refresh_rate is not None) and (refresh_cron is not None):
            raise RuntimeError(
                f"Only one of SOVEREIGN_CONTEXT_REFRESH_RATE or SOVEREIGN_CONTEXT_REFRESH_CRON can be defined. Got {refresh_rate=} and {refresh_cron=}"
            )
        return values

    @root_validator
    def set_default_refresh_rate(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        refresh_rate = values.get("refresh_rate")
        refresh_cron = values.get("refresh_cron")

        if (refresh_rate is None) and (refresh_cron is None):
            values["refresh_rate"] = 3600
        return values

    @validator("refresh_cron")
    def validate_refresh_cron(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if not croniter.is_valid(v):
            raise CroniterBadCronError(f"'{v}' is not a valid cron expression")
        return v

    class Config:
        fields = {
            "refresh": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
            "refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
            "refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
            "refresh_num_retries": {"env": "SOVEREIGN_CONTEXT_REFRESH_NUM_RETRIES"},
            "refresh_retry_interval_secs": {
                "env": "SOVEREIGN_CONTEXT_REFRESH_RETRY_INTERVAL_SECS"
            },
        }


class SourcesConfiguration(BaseSettings):
    refresh_rate: int = 30
    cache_strategy: CacheStrategy = CacheStrategy.context

    class Config:
        fields = {
            "refresh_rate": {"env": "SOVEREIGN_SOURCES_REFRESH_RATE"},
            "cache_strategy": {"env": "SOVEREIGN_CACHE_STRATEGY"},
        }


class LegacyConfig(BaseSettings):
    regions: Optional[List[str]] = None
    eds_priority_matrix: Optional[Dict[str, Dict[str, str]]] = None
    dns_hard_fail: Optional[bool] = None
    environment: Optional[str] = None

    @validator("regions")
    def regions_is_set(cls, v: Optional[List[str]]) -> List[str]:
        if v is not None:
            warnings.warn(
                "Setting regions via config is deprecated. "
                "It is suggested to use a modifier or template "
                "logic in order to achieve the same goal.",
                DeprecationWarning,
            )
            return v
        else:
            return []

    @validator("eds_priority_matrix")
    def eds_priority_matrix_is_set(
        cls, v: Optional[Dict[str, Dict[str, Any]]]
    ) -> Dict[str, Dict[str, Any]]:
        if v is not None:
            warnings.warn(
                "Setting eds_priority_matrix via config is deprecated. "
                "It is suggested to use a modifier or template "
                "logic in order to achieve the same goal.",
                DeprecationWarning,
            )
            return v
        else:
            return {}

    @validator("dns_hard_fail")
    def dns_hard_fail_is_set(cls, v: Optional[bool]) -> bool:
        if v is not None:
            warnings.warn(
                "Setting dns_hard_fail via config is deprecated. "
                "It is suggested to supply a module that can perform "
                "dns resolution to template_context, so that it can "
                "be used via templates instead.",
                DeprecationWarning,
            )
            return v
        else:
            return False

    @validator("environment")
    def environment_is_set(cls, v: Optional[str]) -> Optional[str]:
        if v is not None:
            warnings.warn(
                "Setting environment via config is deprecated. "
                "It is suggested to configure this value through log_fmt "
                "instead.",
                DeprecationWarning,
            )
            return v
        else:
            return None

    class Config:
        fields = {
            "dns_hard_fail": {"env": "SOVEREIGN_DNS_HARD_FAIL"},
            "environment": {"env": "SOVEREIGN_ENVIRONMENT"},
        }


class SovereignConfigv2(BaseSettings):
    sources: List[ConfiguredSource]
    templates: Dict[str, List[TemplateSpecification]]
    source_config: SourcesConfiguration = SourcesConfiguration()
    modifiers: List[str] = []
    global_modifiers: List[str] = []
    template_context: ContextConfiguration = ContextConfiguration()
    matching: NodeMatching = NodeMatching()
    authentication: AuthConfiguration = AuthConfiguration()
    logging: LoggingConfiguration = LoggingConfiguration()
    statsd: StatsdConfig = StatsdConfig()
    sentry_dsn: SecretStr = SecretStr("")
    debug: bool = False
    legacy_fields: LegacyConfig = LegacyConfig()
    discovery_cache: DiscoveryCacheConfig = DiscoveryCacheConfig()

    class Config:
        fields = {
            "sentry_dsn": {"env": "SOVEREIGN_SENTRY_DSN"},
            "debug": {"env": "SOVEREIGN_DEBUG"},
        }

    @property
    def passwords(self) -> List[str]:
        return self.authentication.auth_passwords.get_secret_value().split(",") or []

    def xds_templates(self) -> Dict[str, Dict[str, XdsTemplate]]:
        ret: Dict[str, Dict[str, XdsTemplate]] = {
            "__any__": {}
        }  # Special key to hold templates from all versions
        for version, template_specs in self.templates.items():
            loaded_templates = {
                template.type: XdsTemplate(path=template.spec)
                for template in template_specs
            }
            ret[str(version)] = loaded_templates
            ret["__any__"].update(loaded_templates)
        return ret

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"SovereignConfigv2({self.dict()})"

    def show(self) -> Dict[str, Any]:
        return self.dict()

    @staticmethod
    def from_legacy_config(other: SovereignConfig) -> "SovereignConfigv2":
        new_templates = dict()
        for version, templates in other.templates.items():
            specs = list()
            for type, path in templates.items():
                if isinstance(path, str):
                    specs.append(
                        TemplateSpecification(
                            type=type, spec=Loadable.from_legacy_fmt(path)
                        )
                    )
                else:
                    # Just in case? Although this shouldn't happen
                    specs.append(TemplateSpecification(type=type, spec=path))
            new_templates[str(version)] = specs

        return SovereignConfigv2(
            sources=other.sources,
            templates=new_templates,
            source_config=SourcesConfiguration(
                refresh_rate=other.sources_refresh_rate,
                cache_strategy=CacheStrategy(other.cache_strategy),
            ),
            modifiers=other.modifiers,
            global_modifiers=other.global_modifiers,
            template_context=ContextConfiguration(
                context=ContextConfiguration.context_from_legacy(
                    other.template_context
                ),
                refresh=other.refresh_context,
                refresh_rate=other.context_refresh_rate,
                refresh_cron=other.context_refresh_cron,
            ),
            matching=NodeMatching(
                enabled=other.node_matching,
                source_key=other.source_match_key,
                node_key=other.node_match_key,
            ),
            authentication=AuthConfiguration(
                enabled=other.auth_enabled,
                auth_passwords=SecretStr(other.auth_passwords),
                encryption_key=SecretStr(other.encryption_key),
            ),
            logging=LoggingConfiguration(
                application_logs=ApplicationLogConfiguration(
                    enabled=other.enable_application_logs,
                ),
                access_logs=AccessLogConfiguration(
                    enabled=other.enable_access_logs,
                    log_fmt=other.log_fmt,
                    ignore_empty_fields=other.ignore_empty_log_fields,
                ),
            ),
            statsd=other.statsd,
            sentry_dsn=SecretStr(other.sentry_dsn),
            debug=other.debug_enabled,
            legacy_fields=LegacyConfig(
                regions=other.regions,
                eds_priority_matrix=other.eds_priority_matrix,
                dns_hard_fail=other.dns_hard_fail,
                environment=other.environment,
            ),
            discovery_cache=other.discovery_cache,
        )



import json
import threading
from abc import ABC, abstractmethod
from functools import cached_property
from typing import Any, Dict, Optional

from structlog.exceptions import DropEvent
from structlog.stdlib import BoundLogger

from sovereign.logging.types import EventDict, ProcessedMessage

LOG_QUEUE = threading.local()


class BaseLogger(ABC):
    _user_log_fmt: Optional[str]

    @property
    @abstractmethod
    def is_enabled(self) -> bool:
        ...

    @property
    @abstractmethod
    def _default_log_fmt(self) -> Dict[str, str]:
        ...

    def is_enabled_processor(
        self, logger: BoundLogger, method_name: str, event_dict: EventDict
    ) -> ProcessedMessage:
        if not self.is_enabled:
            raise DropEvent
        return event_dict

    @cached_property
    def get_configured_log_format(self) -> Dict[str, str]:
        if isinstance(self._user_log_fmt, str) and self._user_log_fmt != "":
            format = json.loads(self._user_log_fmt)
            if not isinstance(format, dict):
                raise RuntimeError(
                    f"Failed to parse log format as JSON: {self._user_log_fmt}"
                )
            if "event" not in format:
                format["event"] = "{event}"
            return format
        return self._default_log_fmt

    def merge_in_threadlocal(
        self, logger: Any, method_name: str, event_dict: EventDict
    ) -> ProcessedMessage:
        self._ensure_threadlocal()
        fields: Dict[str, Any] = LOG_QUEUE.fields.copy()
        fields.update(event_dict)
        return fields

    def clear_log_fields(self) -> None:
        LOG_QUEUE.fields = dict()

    def _ensure_threadlocal(self) -> None:
        if not hasattr(LOG_QUEUE, "fields"):
            LOG_QUEUE.fields = dict()

    def queue_log_fields(self, **kwargs: Any) -> None:
        self._ensure_threadlocal()
        LOG_QUEUE.fields.update(kwargs)



# type: ignore
# pylint: disable=no-name-in-module,too-many-branches
""" Stolen from the saltstack library """
from collections.abc import Mapping
import copy


def update(dest, upd, recursive_update=True, merge_lists=False):
    if (not isinstance(dest, Mapping)) or (not isinstance(upd, Mapping)):
        raise TypeError("Cannot update using non-dict types in dictupdate.update()")
    updkeys = list(upd.keys())
    if not set(list(dest.keys())) & set(updkeys):
        recursive_update = False
    if recursive_update:
        for key in updkeys:
            val = upd[key]
            try:
                dest_subkey = dest.get(key, None)
            except AttributeError:
                dest_subkey = None
            if isinstance(dest_subkey, Mapping) and isinstance(val, Mapping):
                ret = update(dest_subkey, val, merge_lists=merge_lists)
                dest[key] = ret
            elif isinstance(dest_subkey, list) and isinstance(val, list):
                if merge_lists:
                    merged = copy.deepcopy(dest_subkey)
                    merged.extend([x for x in val if x not in merged])
                    dest[key] = merged
                else:
                    dest[key] = upd[key]
            else:
                dest[key] = upd[key]
        return dest
    else:
        try:
            for k in upd:
                dest[k] = upd[k]
        except AttributeError:
            # this mapping is not a dict
            for k in upd:
                dest[k] = upd[k]
        return dest


def merge_list(obj_a, obj_b):
    ret = {}
    for key, val in obj_a.items():
        if key in obj_b:
            ret[key] = [val, obj_b[key]]
        else:
            ret[key] = val
    return ret


def merge_recurse(obj_a, obj_b, merge_lists=False):
    copied = copy.deepcopy(obj_a)
    return update(copied, obj_b, merge_lists=merge_lists)


def merge_overwrite(obj_a, obj_b, merge_lists=False):
    for obj in obj_b:
        if obj in obj_a:
            obj_a[obj] = obj_b[obj]
    return merge_recurse(obj_a, obj_b, merge_lists=merge_lists)


def merge(obj_a, obj_b, strategy="recurse", merge_lists=False):
    if strategy == "list":
        merged = merge_list(obj_a, obj_b)
    elif strategy == "recurse":
        merged = merge_recurse(obj_a, obj_b, merge_lists)
    elif strategy == "overwrite":
        merged = merge_overwrite(obj_a, obj_b, merge_lists)
    elif strategy == "none":
        merged = merge_recurse(obj_a, obj_b)
    else:
        merged = merge_recurse(obj_a, obj_b)

    return merged



from typing import Any, Dict
from sovereign.sources.lib import Source
from sovereign.config_loader import Loadable


class File(Source):
    def __init__(self, config: Dict[str, Any], scope: str = "default"):
        super(File, self).__init__(config, scope)
        try:
            self.path = Loadable.from_legacy_fmt(config["path"])
        except KeyError:
            try:
                self.path = Loadable(**config["spec"])
            except KeyError:
                raise KeyError('File source needs to specify "spec" within config')

    def get(self) -> Any:
        """
        Uses the file config loader to load the given path
        """
        return self.path.load()



from typing import Optional, Dict, List
from random import randint
from sovereign.schemas import DiscoveryRequest, Node, Locality, Resources, Status


def mock_discovery_request(
    service_cluster: Optional[str] = None,
    resource_names: Optional[List[str]] = None,
    region: str = "none",
    version: str = "1.11.1",
    metadata: Optional[Dict[str, str]] = None,
    error_message: Optional[str] = None,
) -> DiscoveryRequest:
    if resource_names is None:
        resource_names = Resources()
    else:
        resource_names = Resources(resource_names)
    request = DiscoveryRequest(
        type_url=None,
        node=Node(
            id="mock",
            cluster=service_cluster or "",
            build_version=f"e5f864a82d4f27110359daa2fbdcb12d99e415b9/{version}/Clean/RELEASE",
            locality=Locality(zone=region, region="example", sub_zone="a"),
        ),
        version_info=str(randint(100000, 1000000000)),
        resource_names=resource_names,
        hide_private_keys=True,
        desired_controlplane="controlplane",
        error_detail=Status(code=200, message="None", details=["None"]),
    )
    if isinstance(metadata, dict):
        request.node.metadata = metadata
    if error_message:
        request.error_detail = Status(code=666, message=error_message, details=["foo"])
    return request



from typing import Any, Dict, Optional

from fastapi import APIRouter, Body
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field

from sovereign import json_response_class, logs, server_cipher_container
from sovereign.schemas import EncryptionConfig
from sovereign.utils.crypto.crypto import CipherContainer
from sovereign.utils.crypto.suites import EncryptionType

router = APIRouter()


class EncryptionRequest(BaseModel):
    data: str = Field(..., title="Text to be encrypted", min_length=1, max_length=65535)
    key: Optional[str] = Field(
        None,
        title="Optional encryption key to use to encrypt",
        min_length=44,
        max_length=44,
    )
    encryption_type: str = Field(default="fernet", title="Encryption type to be used")


class DecryptionRequest(BaseModel):
    data: str = Field(..., title="Text to be decrypted", min_length=1, max_length=65535)
    key: str = Field(
        ...,
        title="Encryption key to use to decrypt",
        min_length=44,
        max_length=44,
    )
    encryption_type: str = Field(default="fernet", title="Encryption type to be used")


class DecryptableRequest(BaseModel):
    data: str = Field(..., title="Text to be decrypted", min_length=1, max_length=65535)


@router.post(
    "/decrypt",
    summary="Decrypt provided encrypted data using a provided key",
    response_class=json_response_class,
)
async def _decrypt(request: DecryptionRequest = Body(None)) -> dict[str, Any]:
    user_cipher_container = CipherContainer.from_encryption_configs(
        encryption_configs=[
            EncryptionConfig(
                encryption_key=request.key,
                encryption_type=EncryptionType(request.encryption_type),
            )
        ],
        logger=logs.application_logger.logger,
    )
    return {**user_cipher_container.decrypt_with_type(request.data)}


@router.post(
    "/encrypt",
    summary="Encrypt provided data using this servers key or provided key",
    response_class=json_response_class,
)
async def _encrypt(request: EncryptionRequest = Body(None)) -> dict[str, Any]:
    if request.key:
        user_cipher_container = CipherContainer.from_encryption_configs(
            encryption_configs=[
                EncryptionConfig(
                    encryption_key=request.key,
                    encryption_type=EncryptionType(request.encryption_type),
                )
            ],
            logger=logs.application_logger.logger,
        )
        return {**user_cipher_container.encrypt(request.data)}
    return {**server_cipher_container.encrypt(request.data)}


@router.post(
    "/decryptable",
    summary="Check whether data is decryptable by this server",
    response_class=json_response_class,
)
async def _decryptable(request: DecryptableRequest = Body(None)) -> JSONResponse:
    server_cipher_container.decrypt(request.data)
    return json_response_class({})


@router.get(
    "/generate_key",
    summary="Generate a new asymmetric encryption key",
    response_class=json_response_class,
)
def _generate_key(encryption_type: str = "fernet") -> Dict[str, str]:
    cipher_suite = CipherContainer.get_cipher_suite(EncryptionType(encryption_type))
    return {
        "key": cipher_suite.generate_key().decode(),
        "encryption_type": encryption_type,
    }



from typing import Literal, Self, Sequence

from cachetools import TTLCache, cached
from fastapi.exceptions import HTTPException
from structlog.stdlib import BoundLogger
from typing_extensions import TypedDict

from sovereign.schemas import EncryptionConfig
from sovereign.utils.crypto.suites import (
    AESGCMCipher,
    CipherSuite,
    DisabledCipher,
    EncryptionType,
    FernetCipher,
)


class EncryptOutput(TypedDict):
    encrypted_data: str
    encryption_type: str


class DecryptOutput(TypedDict):
    decrypted_data: str
    encryption_type: str


class CipherContainer:
    """
    Object which intercepts encrypt/decryptions
    Tries to decrypt data using the ciphers provided in order
    Encrypts with the first suite available.
    """

    def __init__(self, suites: Sequence[CipherSuite], logger: BoundLogger) -> None:
        self.suites: Sequence[CipherSuite] = suites
        self.logger = logger

    def encrypt(self, data: str) -> EncryptOutput:
        try:
            # Use the first cipher suite to encrypt the data
            encrypted_data = self.suites[0].encrypt(data)
            return {
                "encrypted_data": encrypted_data,
                "encryption_type": str(self.suites[0]),
            }
        # pylint: disable=broad-except
        except Exception as e:
            self.logger.exception(str(e))
            # TODO: defer this http error to later, return a normal error here
            raise HTTPException(status_code=400, detail="Encryption failed")

    def decrypt_with_type(self, data: str) -> DecryptOutput:
        return self._decrypt(data)

    def decrypt(self, data: str) -> str:
        return self._decrypt(data)["decrypted_data"]

    @cached(cache=TTLCache(maxsize=128, ttl=600))
    def _decrypt(self, data: str) -> DecryptOutput:
        try:
            for suite in self.suites:
                try:
                    decrypted_data = suite.decrypt(data)
                    return {
                        "decrypted_data": decrypted_data,
                        "encryption_type": str(suite),
                    }
                # pylint: disable=broad-except
                except Exception as e:
                    self.logger.exception(str(e))
                    self.logger.debug(f"Failed to decrypt with suite {suite}")
            else:
                raise ValueError("Could not decrypt with any suite")
        except Exception as e:
            self.logger.exception(str(e))
            # TODO: defer this http error to later, return a normal error here
            raise HTTPException(status_code=400, detail="Decryption failed")

    @property
    def key_available(self) -> bool:
        return self.suites[0].key_available

    AVAILABLE_CIPHERS: dict[EncryptionType | Literal["default"], type[CipherSuite]] = {
        EncryptionType.DISABLED: DisabledCipher,
        EncryptionType.AESGCM: AESGCMCipher,
        EncryptionType.FERNET: FernetCipher,
        "default": FernetCipher,
    }

    @classmethod
    def get_cipher_suite(cls, encryption_type: EncryptionType) -> type[CipherSuite]:
        SelectedCipher = cls.AVAILABLE_CIPHERS.get(
            encryption_type, cls.AVAILABLE_CIPHERS["default"]
        )
        return SelectedCipher

    @classmethod
    def create_cipher_suite(
        cls,
        encryption_type: EncryptionType,
        key: str,
        logger: BoundLogger,
    ) -> CipherSuite:
        kwargs = {
            "secret_key": key,
        }
        try:
            SelectedCipher = cls.get_cipher_suite(encryption_type)
            return SelectedCipher(**kwargs)
        except TypeError:
            pass
        except ValueError as e:
            if key not in (b"", ""):
                logger.error(
                    f"Encryption key was provided, but appears to be invalid: {repr(e)}"
                )
        return DisabledCipher(**kwargs)

    @classmethod
    def from_encryption_configs(
        cls, encryption_configs: Sequence[EncryptionConfig], logger: BoundLogger
    ) -> Self:
        cipher_suites: list[CipherSuite] = []
        for encryption_config in encryption_configs:
            cipher_suites.append(
                cls.create_cipher_suite(
                    key=encryption_config.encryption_key,
                    encryption_type=encryption_config.encryption_type,
                    logger=logger,
                )
            )
        return cls(suites=cipher_suites, logger=logger)



import os
from typing import Any, Mapping

from pydantic.error_wrappers import ValidationError

from sovereign import config_loader
from sovereign.context import TemplateContext
from sovereign.logging.bootstrapper import LoggerBootstrapper
from sovereign.schemas import SovereignAsgiConfig, SovereignConfig, SovereignConfigv2
from sovereign.sources import SourcePoller
from sovereign.statistics import configure_statsd
from sovereign.utils.crypto.crypto import CipherContainer
from sovereign.utils.dictupdate import merge  # type: ignore


def parse_raw_configuration(path: str) -> Mapping[Any, Any]:
    ret: Mapping[Any, Any] = dict()
    for p in path.split(","):
        spec = config_loader.Loadable.from_legacy_fmt(p)
        ret = merge(obj_a=ret, obj_b=spec.load(), merge_lists=True)
    return ret


def load_sovereign_configuration() -> SovereignConfigv2:
    config_path = os.getenv("SOVEREIGN_CONFIG", "file:///etc/sovereign.yaml")
    try:
        return SovereignConfigv2(**parse_raw_configuration(config_path))
    except ValidationError:
        old_config = SovereignConfig(**parse_raw_configuration(config_path))
        return SovereignConfigv2.from_legacy_config(old_config)


CONFIG = load_sovereign_configuration()
ASGI_CONFIG = SovereignAsgiConfig()
XDS_TEMPLATES = CONFIG.xds_templates()

LOGS = LoggerBootstrapper(CONFIG)
STATS = configure_statsd(config=CONFIG.statsd)
ENCRYPTION_CONFIGS = CONFIG.authentication.encryption_configs
CIPHER_CONTAINER = CipherContainer.from_encryption_configs(
    encryption_configs=ENCRYPTION_CONFIGS,
    logger=LOGS.application_logger.logger,
)

POLLER = SourcePoller(
    sources=CONFIG.sources,
    matching_enabled=CONFIG.matching.enabled,
    node_match_key=CONFIG.matching.node_key,
    source_match_key=CONFIG.matching.source_key,
    source_refresh_rate=CONFIG.source_config.refresh_rate,
    logger=LOGS.application_logger.logger,
    stats=STATS,
)
TEMPLATE_CONTEXT = TemplateContext(
    refresh_rate=CONFIG.template_context.refresh_rate,
    refresh_cron=CONFIG.template_context.refresh_cron,
    refresh_num_retries=CONFIG.template_context.refresh_num_retries,
    refresh_retry_interval_secs=CONFIG.template_context.refresh_retry_interval_secs,
    configured_context=CONFIG.template_context.context,
    poller=POLLER,
    encryption_suite=CIPHER_CONTAINER,
    logger=LOGS.application_logger.logger,
    stats=STATS,
)
POLLER.lazy_load_modifiers(CONFIG.modifiers)
POLLER.lazy_load_global_modifiers(CONFIG.global_modifiers)



"""
Modifier abstract class
'''''''''''''''''''''''
This class can be subclassed, installed as an entry point, and then
used via configuration.

`todo entry point install guide`
"""
import abc
from typing import List, Any, Dict


class Modifier(metaclass=abc.ABCMeta):
    """
    Modifier is an abstract base class used to change instances in-flight.

    :param instance: A single instance obtained from any source
    :type instance: dict
    """

    def __init__(self, instance: Dict[str, Any]) -> None:
        self.instance = instance

    @abc.abstractmethod
    def match(self) -> bool:
        """
        match is an abstract method which must be overwritten by all
        inheriting classes.
        This is run prior to applying a modifier, to ensure that it's
        being applied to the correct object.
        Match must return something truthy or falsy.
        """

    @abc.abstractmethod
    def apply(self) -> None:
        """
        apply is an abstract method which must be overwritten by all
        inheriting classes.
        Apply should modify a `self.instance` object in-place.
        """


class GlobalModifier:
    """
    GlobalModifier is an abstract base class used to change instance data in-flight.

    :param source_data: A list of instances obtained from any source
    :type source_data: list
    """

    def __init__(self, source_data: List[Dict[str, Any]]) -> None:
        self.data = source_data
        self.unmatched: List[Dict[str, Any]] = list()
        self.matched: List[Dict[str, Any]] = list()
        self._match()

    @abc.abstractmethod
    def match(self, data_instance: Dict[str, Any]) -> bool:
        """
        match is an abstract method which must be overwritten by all
        inheriting classes.
        This is run prior to applying a global modifier, and results in
        the given data source being sorted into 'matched' and 'unmatched'
        groups.
        Match must return something truthy or falsy.

        :param data_instance: dict object to be matched against
        :return: True if matched, or False if unmatched
        """

    def _match(self) -> None:
        """
        Sorts the given data into two tuples, matched & unmatched, using
        the self.match method
        """
        self.matched = [i for i in self.data if self.match(i)]
        self.unmatched = [i for i in self.data if not self.match(i)]

    @abc.abstractmethod
    def apply(self) -> None:
        """
        apply is an abstract method which must be overwritten by all
        inheriting classes.
        Apply should modify the list object `self.matched` in-place
        """

    def join(self) -> List[Dict[str, Any]]:
        """
        Joins matched and unmatched sets of data back together.
        This is run after the modifier has been applied.
        """
        return self.matched + self.unmatched



from typing import Dict
from functools import cached_property
from importlib.metadata import entry_points, EntryPoints


class EntryPointLoader:
    ENTRY_POINTS = entry_points()

    def __init__(self, *args: str) -> None:
        self._groups = args

    @classmethod
    def _select(cls, group: str) -> EntryPoints:
        return cls.ENTRY_POINTS.select(group=f"sovereign.{group}")

    @cached_property
    def groups(self) -> Dict[str, EntryPoints]:
        return {group: self._select(group) for group in self._groups}



import os
import time
from uuid import uuid4
from fastapi.requests import Request
from fastapi.responses import Response
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from sovereign import config, logs, get_request_id, _request_id_ctx_var, stats


class RequestContextLogMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        response = Response("Internal server error", status_code=500)
        token = _request_id_ctx_var.set(str(uuid4()))
        try:
            response = await call_next(request)
        finally:
            req_id = get_request_id()
            response.headers["X-Request-ID"] = req_id
            logs.access_logger.queue_log_fields(REQUEST_ID=req_id)
            _request_id_ctx_var.reset(token)
        return response


class LoggingMiddleware(BaseHTTPMiddleware):
    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        start_time = time.time()
        response = Response("Internal server error", status_code=500)

        source_ip, source_port = "0.0.0.0", 0
        if addr := request.client:
            source_ip = addr.host
            source_port = addr.port
        if xff := request.headers.get("X-Forwarded-For"):
            source_ip = xff.split(",")[0]  # leftmost address
        logs.access_logger.clear_log_fields()
        logs.access_logger.queue_log_fields(
            ENVIRONMENT=config.legacy_fields.environment,
            HOST=request.headers.get("host", "-"),
            METHOD=request.method,
            PATH=request.url.path,
            QUERY=dict(request.query_params.items()),
            SOURCE_IP=source_ip,
            SOURCE_PORT=source_port,
            PID=os.getpid(),
            USER_AGENT=request.headers.get("user-agent", "-"),
            BYTES_RX=request.headers.get("content-length", "-"),
        )
        try:
            response = await call_next(request)
        finally:
            duration = time.time() - start_time
            logs.access_logger.queue_log_fields(
                BYTES_TX=response.headers.get("content-length", "-"),
                STATUS_CODE=response.status_code,
                DURATION=duration,
            )
            if "discovery" in str(request.url):
                request_info = {
                    "path": request.url.path,
                    "xds_type": response.headers.get("X-Sovereign-Requested-Type"),
                    "client_version": response.headers.get("X-Sovereign-Client-Build"),
                    "response_code": response.status_code,
                }
                tags = [
                    ":".join(map(str, [k, v]))
                    for k, v in request_info.items()
                    if v is not None
                ]
                stats.increment("discovery.rq_total", tags=tags)
                stats.timing("discovery.rq_ms", value=duration * 1000, tags=tags)
            logs.access_logger.logger.info("request")
        return response



import asyncio
import traceback
from copy import deepcopy
from importlib.metadata import EntryPoint
from datetime import timedelta, datetime
from typing import Iterable, Any, Dict, List, Union, Type, Optional

from glom import glom, PathAccessError

from sovereign.utils.entry_point_loader import EntryPointLoader
from sovereign.sources.lib import Source
from sovereign.modifiers.lib import Modifier, GlobalModifier
from sovereign.schemas import (
    ConfiguredSource,
    SourceData,
    Node,
)

from structlog.stdlib import BoundLogger


def is_debug_request(v: str, debug: bool = False) -> bool:
    return v == "" and debug


def is_wildcard(v: List[str]) -> bool:
    return v in [["*"], "*", ("*",)]


def contains(container: Iterable[Any], item: Any) -> bool:
    return item in container


Mods = Dict[str, Type[Modifier]]
GMods = Dict[str, Type[GlobalModifier]]


class SourcePoller:
    def __init__(
        self,
        sources: List[ConfiguredSource],
        matching_enabled: bool,
        node_match_key: str,
        source_match_key: str,
        source_refresh_rate: int,
        logger: BoundLogger,
        stats: Any,
    ):
        self.matching_enabled = matching_enabled
        self.node_match_key = node_match_key
        self.source_match_key = source_match_key
        self.source_refresh_rate = source_refresh_rate
        self.logger = logger
        self.stats = stats

        self.entry_points = EntryPointLoader("sources", "modifiers", "global_modifiers")

        self.source_classes: Dict[str, Type[Source]] = {
            e.name: e.load() for e in self.entry_points.groups["sources"]
        }
        self.sources = [self.setup_source(s) for s in sources]
        if not self.sources:
            raise RuntimeError("No data sources available!")

        # These have to be loaded later to avoid circular imports
        self.modifiers: Mods = dict()
        self.global_modifiers: GMods = dict()

        # initially set data and modify
        self.source_data = self.refresh()
        self.last_updated = datetime.now()
        self.instance_count = 0

    @property
    def data_is_stale(self) -> bool:
        return self.last_updated < datetime.now() - timedelta(minutes=2)

    def setup_source(self, configured_source: ConfiguredSource) -> Source:
        source_class = self.source_classes[configured_source.type]
        source = source_class(
            config=configured_source.config,
            scope=configured_source.scope,
        )
        source.setup()
        return source

    def lazy_load_modifiers(self, modifiers: List[str]) -> None:
        if len(self.modifiers) == len(modifiers):
            return
        self.modifiers = self.load_modifier_entrypoints(
            self.entry_points.groups["modifiers"], modifiers
        )

    def lazy_load_global_modifiers(self, global_modifiers: List[str]) -> None:
        if len(self.global_modifiers) == len(global_modifiers):
            return
        self.global_modifiers = self.load_global_modifier_entrypoints(
            self.entry_points.groups["global_modifiers"], global_modifiers
        )

    def load_modifier_entrypoints(
        self, entry_points: Iterable[EntryPoint], configured_modifiers: List[str]
    ) -> Dict[str, Type[Modifier]]:
        ret = dict()
        for entry_point in entry_points:
            if entry_point.name in configured_modifiers:
                self.logger.info(f"Loading modifier {entry_point.name}")
                ret[entry_point.name] = entry_point.load()
        loaded = len(ret)
        configured = len(configured_modifiers)
        assert loaded == configured, (
            f"Number of modifiers loaded ({loaded})"
            f"differ from configured: {configured_modifiers}"
        )
        return ret

    def load_global_modifier_entrypoints(
        self, entry_points: Iterable[EntryPoint], configured_modifiers: List[str]
    ) -> Dict[str, Type[GlobalModifier]]:
        ret = dict()
        for entry_point in entry_points:
            if entry_point.name in configured_modifiers:
                self.logger.info(f"Loading global modifier {entry_point.name}")
                ret[entry_point.name] = entry_point.load()

        loaded = len(ret)
        configured = len(configured_modifiers)
        assert loaded == configured, (
            f"Number of global modifiers loaded ({loaded})"
            f"differ from configured: {configured_modifiers}"
        )
        return ret

    def apply_modifications(self, data: Optional[SourceData]) -> SourceData:
        if data is None:
            data = self.source_data
        if len(self.modifiers) or len(self.global_modifiers):
            with self.stats.timed("modifiers.apply_ms"):
                data = deepcopy(data)
                for scope, instances in data.scopes.items():
                    for g in self.global_modifiers.values():
                        global_modifier = g(instances)
                        global_modifier.apply()
                        data.scopes[scope] = global_modifier.join()

                    for instance in data.scopes[scope]:
                        for m in self.modifiers.values():
                            modifier = m(instance)
                            if modifier.match():
                                # Modifies the instance in-place
                                modifier.apply()
        return data

    def refresh(self) -> SourceData:
        self.stats.increment("sources.attempt")
        try:
            new = SourceData()
            for source in self.sources:
                new.scopes[source.scope].extend(source.get())
        except Exception as e:
            self.logger.error(
                event="Error while refreshing sources",
                traceback=[line for line in traceback.format_exc().split("\n")],
                error=e.__class__.__name__,
                detail=getattr(e, "detail", "-"),
            )
            self.stats.increment("sources.error")
            raise

        # Is the new data the same as what we currently have
        if new == getattr(self, "source_data", None):
            self.stats.increment("sources.unchanged")
            self.last_updated = datetime.now()
            return self.source_data
        else:
            self.stats.increment("sources.refreshed")
            self.last_updated = datetime.now()
            self.instance_count = len(
                [instance for scope in new.scopes.values() for instance in scope]
            )
            return new

    def extract_node_key(self, node: Union[Node, Dict[Any, Any]]) -> Any:
        if "." not in self.node_match_key:
            # key is not nested, don't need glom
            node_value = getattr(node, self.node_match_key)
        else:
            try:
                node_value = glom(node, self.node_match_key)
            except PathAccessError:
                raise RuntimeError(
                    f'Failed to find key "{self.node_match_key}" in discoveryRequest({node}).\n'
                    f"See the docs for more info: "
                    f"https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html"
                )
        return node_value

    def extract_source_key(self, source: Dict[Any, Any]) -> Any:
        if "." not in self.source_match_key:
            # key is not nested, don't need glom
            source_value = source[self.source_match_key]
        else:
            try:
                source_value = glom(source, self.source_match_key)
            except PathAccessError:
                raise RuntimeError(
                    f'Failed to find key "{self.source_match_key}" in instance({source}).\n'
                    f"See the docs for more info: "
                    f"https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/node_matching.html"
                )
        return source_value

    def match_node(
        self,
        node_value: Any,
        modify: bool = True,
    ) -> SourceData:
        """
        Checks a node against all sources, using the node_match_key and source_match_key
        to determine if the node should receive the source in its configuration.
        """

        if self.data_is_stale:
            # Log/emit metric and manually refresh sources.
            self.stats.increment("sources.stale")
            self.logger.debug(
                "Sources have not been refreshed in 2 minutes",
                last_update=self.last_updated,
                instance_count=self.instance_count,
            )
            self.poll()

        ret = SourceData()

        if modify:
            if not hasattr(self, "source_data_modified"):
                self.poll()
            data = self.source_data_modified
        else:
            data = self.source_data

        for scope, instances in data.scopes.items():
            if self.matching_enabled is False:
                ret.scopes[scope] = instances
                continue

            for instance in instances:
                source_value = self.extract_source_key(instance)

                # If a single expression evaluates true, the remaining are not evaluated/executed.
                # This saves (a small amount of) computation, which helps when the server starts
                # to receive thousands of requests. The list has been ordered descending by what
                # we think will more commonly be true.
                match = (
                    contains(source_value, node_value)
                    or node_value == source_value
                    or is_wildcard(node_value)
                    or is_wildcard(source_value)
                    or is_debug_request(node_value)
                )
                if match:
                    ret.scopes[scope].append(instance)
        return ret

    @property
    def match_keys(self) -> List[str]:
        """
        Checks for all match keys present in existing sources and adds them to a list

        A dict is used instead of a set because dicts cannot have duplicate keys, and
        have ordering since python 3.6
        """
        ret: Dict[str, None] = dict()
        ret["*"] = None
        for _, instances in self.source_data.scopes.items():
            if self.matching_enabled is False:
                break
            for instance in instances:
                source_value = glom(instance, self.source_match_key)
                if isinstance(source_value, str):
                    ret[source_value] = None
                elif isinstance(source_value, Iterable):
                    for item in source_value:
                        ret[item] = None
                    continue
                ret[source_value] = None
        return list(ret.keys())

    def poll(self) -> None:
        self.source_data = self.refresh()
        self.source_data_modified = self.apply_modifications(self.source_data)

    async def poll_forever(self) -> None:
        while True:
            self.poll()
            await asyncio.sleep(self.source_refresh_rate)



from typing import Dict

from fastapi import Body, Header
from fastapi.responses import Response
from fastapi.routing import APIRouter

from sovereign import config, discovery, logs
from sovereign.schemas import DiscoveryRequest, DiscoveryResponse, ProcessedTemplate
from sovereign.utils.auth import authenticate
from sovereign.utils.version_info import compute_hash

discovery_cache = config.discovery_cache

if discovery_cache.enabled:
    from cashews import cache

    cache.setup(
        f"{discovery_cache.protocol}{discovery_cache.host}:{discovery_cache.port}",
        password=discovery_cache.password.get_secret_value(),
        client_side=discovery_cache.client_side,
        wait_for_connection_timeout=discovery_cache.wait_for_connection_timeout,
        socket_connect_timeout=discovery_cache.socket_connect_timeout,
        socket_timeout=discovery_cache.socket_timeout,
        max_connections=discovery_cache.max_connections,
        retry_on_timeout=discovery_cache.retry_on_timeout,
        suppress=discovery_cache.suppress,
        socket_keepalive=discovery_cache.socket_keepalive,
        enable=discovery_cache.enabled,
    )

router = APIRouter()

type_urls = {
    "v2": {
        "listeners": "type.googleapis.com/envoy.api.v2.Listener",
        "clusters": "type.googleapis.com/envoy.api.v2.Cluster",
        "endpoints": "type.googleapis.com/envoy.api.v2.ClusterLoadAssignment",
        "secrets": "type.googleapis.com/envoy.api.v2.auth.Secret",
        "routes": "type.googleapis.com/envoy.api.v2.RouteConfiguration",
        "scoped-routes": "type.googleapis.com/envoy.api.v2.ScopedRouteConfiguration",
    },
    "v3": {
        "listeners": "type.googleapis.com/envoy.config.listener.v3.Listener",
        "clusters": "type.googleapis.com/envoy.config.cluster.v3.Cluster",
        "endpoints": "type.googleapis.com/envoy.config.endpoint.v3.ClusterLoadAssignment",
        "secrets": "type.googleapis.com/envoy.extensions.transport_sockets.tls.v3.Secret",
        "routes": "type.googleapis.com/envoy.config.route.v3.RouteConfiguration",
        "scoped-routes": "type.googleapis.com/envoy.config.route.v3.ScopedRouteConfiguration",
        "runtime": "type.googleapis.com/envoy.service.runtime.v3.Runtime",
    },
}


def response_headers(
    discovery_request: DiscoveryRequest, response: ProcessedTemplate, xds: str
) -> Dict[str, str]:
    return {
        "X-Sovereign-Client-Build": discovery_request.envoy_version,
        "X-Sovereign-Client-Version": discovery_request.version_info,
        "X-Sovereign-Requested-Resources": ",".join(discovery_request.resource_names)
        or "all",
        "X-Sovereign-Requested-Type": xds,
        "X-Sovereign-Response-Version": response.version,
    }


@router.post(
    "/{version}/discovery:{xds_type}",
    summary="Envoy Discovery Service Endpoint",
    response_model=DiscoveryResponse,
    responses={
        200: {"description": "New resources provided"},
        304: {"description": "Resources are up-to-date"},
        404: {"description": "No resources found"},
    },
)
async def discovery_response(
    version: str,
    xds_type: str,
    discovery_request: DiscoveryRequest = Body(...),
    host: str = Header("no_host_provided"),
) -> Response:
    discovery_request.desired_controlplane = host
    response = await perform_discovery(
        discovery_request, version, xds_type, skip_auth=False
    )
    logs.access_logger.queue_log_fields(
        XDS_RESOURCES=discovery_request.resource_names,
        XDS_ENVOY_VERSION=discovery_request.envoy_version,
        XDS_CLIENT_VERSION=discovery_request.version_info,
        XDS_SERVER_VERSION=response.version,
    )
    if discovery_request.error_detail:
        logs.access_logger.queue_log_fields(
            XDS_ERROR_DETAIL=discovery_request.error_detail.message
        )
    headers = response_headers(discovery_request, response, xds_type)

    if response.version == discovery_request.version_info:
        return not_modified(headers)
    elif getattr(response, "resources", None) == []:
        return Response(status_code=404, headers=headers)
    elif response.version != discovery_request.version_info:
        return Response(
            response.rendered, headers=headers, media_type="application/json"
        )
    return Response(content="Resources could not be determined", status_code=500)


async def perform_discovery(
    req: DiscoveryRequest,
    api_version: str,
    resource_type: str,
    skip_auth: bool = False,
) -> ProcessedTemplate:
    if not skip_auth:
        authenticate(req)
    if discovery_cache.enabled:
        logs.access_logger.queue_log_fields(CACHE_XDS_HIT=False)
        cache_key = compute_hash(
            [
                api_version,
                resource_type,
                req.envoy_version,
                req.resource_names,
                req.desired_controlplane,
                req.hide_private_keys,
                req.type_url,
                req.node.cluster,
                req.node.locality,
                req.node.metadata.get("auth", None),
                req.node.metadata.get("num_cpus", None),
            ]
        )
        if template := await cache.get(key=cache_key, default=None):
            logs.access_logger.queue_log_fields(CACHE_XDS_HIT=True)
            return template  # type: ignore[no-any-return]
    template = discovery.response(req, resource_type)
    type_url = type_urls.get(api_version, {}).get(resource_type)
    if type_url is not None:
        for resource in template.resources:
            if not resource.get("@type"):
                resource["@type"] = type_url
    if discovery_cache.enabled:
        await cache.set(
            key=cache_key,
            value=template,
            expire=discovery_cache.ttl,
        )
    return template


def not_modified(headers: Dict[str, str]) -> Response:
    return Response(status_code=304, headers=headers)



"""
Discovery
---------

Functions used to render and return discovery responses to Envoy proxies.

The templates are configurable. `todo See ref:Configuration#Templates`
"""
from enum import Enum
from typing import List, Dict, Any, Optional

import yaml
from yaml.parser import ParserError, ScannerError  # type: ignore
from starlette.exceptions import HTTPException

try:
    import sentry_sdk

    SENTRY_INSTALLED = True
except ImportError:
    SENTRY_INSTALLED = False

from sovereign import XDS_TEMPLATES, config, logs, template_context
from sovereign.utils.version_info import compute_hash
from sovereign.schemas import XdsTemplate, DiscoveryRequest, ProcessedTemplate


try:
    default_templates = XDS_TEMPLATES["default"]
except KeyError:
    raise KeyError(
        "Your configuration should contain default templates. For more details, see "
        "https://vsyrakis.bitbucket.io/sovereign/docs/html/guides/tutorial.html#create-templates "
    )

cache_strategy = config.source_config.cache_strategy

# Create an enum that bases all the available discovery types off what has been configured
discovery_types = (_type for _type in sorted(XDS_TEMPLATES["__any__"].keys()))
discovery_types_base: Dict[str, str] = {t: t for t in discovery_types}
# TODO: this needs to be typed somehow, but I have no idea how
DiscoveryTypes = Enum("DiscoveryTypes", discovery_types_base)  # type: ignore


def select_template(
    request: DiscoveryRequest,
    discovery_type: str,
    templates: Optional[Dict[str, Dict[str, XdsTemplate]]] = None,
) -> XdsTemplate:
    if templates is None:
        templates = XDS_TEMPLATES
    version = request.envoy_version
    selection = "default"
    for v in templates.keys():
        if version.startswith(v):
            selection = v
    selected_version = templates[selection]
    try:
        resource_type = discovery_type
        return selected_version[resource_type]
    except KeyError:
        raise KeyError(
            f"Unable to get {discovery_type} for template "
            f'version "{selection}". Envoy client version: {version}'
        )


def response(request: DiscoveryRequest, xds_type: str) -> ProcessedTemplate:
    """
    A Discovery **Request** typically looks something like:

    .. code-block:: json

        {
            "version_info": "0",
            "node": {
                "cluster": "T1",
                "build_version": "<revision hash>/<version>/Clean/RELEASE",
                "metadata": {
                    "auth": "..."
                }
            }
        }

    When we receive this, we give the client the latest configuration via a
    Discovery **Response** that looks something like this:

    .. code-block:: json

        {
            "version_info": "abcdef1234567890",
            "resources": []
        }

    The version_info is derived from :func:`sovereign.discovery.version_hash`

    :param request: An envoy Discovery Request
    :param xds_type: what type of XDS template to use when rendering
    :return: An envoy Discovery Response
    """
    template: XdsTemplate = select_template(request, xds_type)
    context = dict(
        discovery_request=request,
        host_header=request.desired_controlplane,
        resource_names=request.resources,
        **template_context.get_context(request, template),
    )
    content = template(**context)

    # Deserialize YAML output from Jinja2
    if not template.is_python_source:
        if not isinstance(content, str):
            raise RuntimeError(
                f"Attempting to deserialize potential non-string data: {content}"
            )
        content = deserialize_config(content)

    # Early return if the template is identical
    config_version = compute_hash(content)
    if config_version == request.version_info and not config.discovery_cache.enabled:
        return ProcessedTemplate(version_info=config_version, resources=[])

    if not isinstance(content, dict):
        raise RuntimeError(f"Attempting to filter unstructured data: {content}")
    resources = filter_resources(content["resources"], request.resources)
    return ProcessedTemplate(resources=resources, version_info=config_version)


def deserialize_config(content: str) -> Dict[str, Any]:
    try:
        envoy_configuration = yaml.safe_load(content)
    except (ParserError, ScannerError) as e:
        logs.access_logger.queue_log_fields(
            error=repr(e),
            YAML_CONTEXT=e.context,
            YAML_CONTEXT_MARK=e.context_mark,
            YAML_NOTE=e.note,
            YAML_PROBLEM=e.problem,
            YAML_PROBLEM_MARK=e.problem_mark,
        )

        if SENTRY_INSTALLED and config.sentry_dsn:
            sentry_sdk.capture_exception(e)

        raise HTTPException(
            status_code=500,
            detail="Failed to load configuration, there may be "
            "a syntax error in the configured templates. "
            "Please check Sentry if you have configured Sentry DSN",
        )
    if not isinstance(envoy_configuration, dict):
        raise RuntimeError(
            f"Deserialized configuration is of unexpected format: {envoy_configuration}"
        )
    return envoy_configuration


def filter_resources(
    generated: List[Dict[str, Any]], requested: List[str]
) -> List[Dict[str, Any]]:
    """
    If Envoy specifically requested a resource, this removes everything
    that does not match the name of the resource.
    If Envoy did not specifically request anything, every resource is retained.
    """
    return [resource for resource in generated if resource_name(resource) in requested]


def resource_name(resource: Dict[str, Any]) -> str:
    name = resource.get("name") or resource.get("cluster_name")
    if isinstance(name, str):
        return name
    raise KeyError(
        f"Failed to determine the name or cluster_name of the following resource: {resource}"
    )



from abc import ABC, abstractmethod
from typing import Any


class CipherSuite(ABC):
    @abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    @abstractmethod
    def encrypt(self, data: str) -> str:
        ...

    @abstractmethod
    def decrypt(self, data: str) -> str:
        ...

    @property
    @abstractmethod
    def key_available(self) -> bool:
        ...

    @classmethod
    @abstractmethod
    def generate_key(cls) -> bytes:
        ...



from functools import cached_property
from typing import Any, Dict

import structlog
from structlog.stdlib import BoundLogger

from sovereign.logging.base_logger import BaseLogger
from sovereign.logging.types import EventDict, LoggingType, ProcessedMessage
from sovereign.schemas import SovereignConfigv2


class ApplicationLogger(BaseLogger):
    def __init__(self, root_logger: BoundLogger, config: SovereignConfigv2):
        self._application_logs_enabled = config.logging.application_logs.enabled
        self._user_log_fmt = config.logging.application_logs.log_fmt

        self.logger: BoundLogger = structlog.wrap_logger(
            root_logger,
            wrapper_class=structlog.BoundLogger,
            processors=[
                self.is_enabled_processor,
                self.merge_in_threadlocal,
                self.format_application_log_fields,
            ],
            type=LoggingType.APPLICATION,
        )

    @cached_property
    def is_enabled(self) -> bool:
        return self._application_logs_enabled

    @cached_property
    def _default_log_fmt(self) -> Dict[str, str]:
        return {
            "type": "{type}",
            "event": "{event}",
            "error": "{error}",
            "traceback": "{traceback}",
            "last_update": "{last_update}",
            "instance_count": "{instance_count}",
        }

    def format_application_log_fields(
        self, logger: BoundLogger, method_name: str, event_dict: EventDict
    ) -> ProcessedMessage:
        formatted_dict: Dict[str, Any] = {
            "level": method_name,
        }
        for k, v in self.get_configured_log_format.items():
            try:
                value: str = v.format(**event_dict)
            except KeyError:
                continue
            formatted_dict[k] = value
        return formatted_dict



from collections import defaultdict
from typing import Any, Dict, List

from fastapi import APIRouter, Cookie, Path, Query
from fastapi.encoders import jsonable_encoder
from fastapi.requests import Request
from fastapi.responses import HTMLResponse, JSONResponse, RedirectResponse, Response

from sovereign import XDS_TEMPLATES, html_templates, json_response_class, poller
from sovereign.discovery import DiscoveryTypes
from sovereign.utils.mock import mock_discovery_request
from sovereign.views.discovery import perform_discovery

router = APIRouter()

all_types = [t.value for t in DiscoveryTypes]


@router.get("/")
async def ui_main(request: Request) -> HTMLResponse:
    try:
        return html_templates.TemplateResponse(
            request=request,
            name="base.html",
            media_type="text/html",
            context={
                "all_types": all_types,
                "last_update": str(poller.last_updated),
            },
        )
    except IndexError:
        return html_templates.TemplateResponse(
            request=request,
            name="err.html",
            media_type="text/html",
            context={
                "title": "No resource types configured",
                "message": "A template should be defined for every resource "
                "type that you want your envoy proxies to discover.",
                "doc_link": "https://vsyrakis.bitbucket.io/sovereign/docs/tutorial/templates/",
            },
        )


@router.get(
    "/set-version", summary="Filter the UI by a certain Envoy Version (stores a Cookie)"
)
async def set_envoy_version(
    request: Request,
    version: str = Query(
        "__any__", title="The clients envoy version to emulate in this XDS request"
    ),
) -> Response:
    url = request.headers.get("Referer", "/ui")
    response = RedirectResponse(url=url)
    response.set_cookie(key="envoy_version", value=version)
    return response


@router.get(
    "/set-service-cluster",
    summary="Filter the UI by a certain service cluster (stores a Cookie)",
)
async def set_service_cluster(
    request: Request,
    service_cluster: str = Query(
        "__any__", title="The clients envoy version to emulate in this XDS request"
    ),
) -> Response:
    url = request.headers.get("Referer", "/ui")
    response = RedirectResponse(url=url)
    response.set_cookie(key="service_cluster", value=service_cluster)
    return response


@router.get(
    "/resources/{xds_type}", summary="List available resources for a given xDS type"
)
async def resources(
    request: Request,
    xds_type: str = Path(title="xDS type", description="The type of request"),
    region: str = Query(
        None, title="The clients region to emulate in this XDS request"
    ),
    api_version: str = Query("v2", title="The desired Envoy API version"),
    service_cluster: str = Cookie(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    envoy_version: str = Cookie(
        "__any__", title="The clients envoy version to emulate in this XDS request"
    ),
) -> HTMLResponse:
    ret: Dict[str, List[Dict[str, Any]]] = defaultdict(list)
    try:
        response = await perform_discovery(
            req=mock_discovery_request(
                service_cluster=service_cluster,
                resource_names=[],
                version=envoy_version,
                region=region,
            ),
            api_version=api_version,
            resource_type=xds_type,
            skip_auth=True,
        )
    except KeyError:
        ret["resources"] = []
    else:
        ret["resources"] += response.deserialize_resources()
    return html_templates.TemplateResponse(
        request=request,
        name="resources.html",
        media_type="text/html",
        context={
            "resources": ret["resources"],
            "resource_type": xds_type,
            "all_types": all_types,
            "version": envoy_version,
            "available_versions": list(XDS_TEMPLATES.keys()),
            "service_cluster": service_cluster,
            "available_service_clusters": poller.match_keys,
            "last_update": str(poller.last_updated),
        },
    )


@router.get(
    "/resources/{xds_type}/{resource_name}",
    summary="Return JSON representation of a resource",
)
async def resource(
    xds_type: str = Path(title="xDS type", description="The type of request"),
    resource_name: str = Path(..., title="Name of the resource to view"),
    region: str = Query(
        None, title="The clients region to emulate in this XDS request"
    ),
    api_version: str = Query("v2", title="The desired Envoy API version"),
    service_cluster: str = Cookie(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    envoy_version: str = Cookie(
        "__any__", title="The clients envoy version to emulate in this XDS request"
    ),
) -> Response:
    response = await perform_discovery(
        req=mock_discovery_request(
            service_cluster=service_cluster,
            resource_names=[resource_name],
            version=envoy_version,
            region=region,
        ),
        api_version=api_version,
        resource_type=xds_type,
        skip_auth=True,
    )
    return Response(response.rendered, media_type="application/json")


@router.get(
    "/resources/routes/{route_configuration}/{virtual_host}",
    summary="Return JSON representation of Virtual Hosts",
)
async def virtual_hosts(
    route_configuration: str = Path(..., title="Name of the route configuration"),
    virtual_host: str = Path(..., title="Name of the resource to view"),
    region: str = Query(
        None, title="The clients region to emulate in this XDS request"
    ),
    api_version: str = Query("v2", title="The desired Envoy API version"),
    service_cluster: str = Cookie(
        "*", title="The clients service cluster to emulate in this XDS request"
    ),
    envoy_version: str = Cookie(
        "__any__", title="The clients envoy version to emulate in this XDS request"
    ),
) -> Response:
    response = await perform_discovery(
        req=mock_discovery_request(
            service_cluster=service_cluster,
            resource_names=[route_configuration],
            version=envoy_version,
            region=region,
        ),
        api_version=api_version,
        resource_type="routes",
        skip_auth=True,
    )
    route_configs = [
        resource_
        for resource_ in response.deserialize_resources()
        if resource_["name"] == route_configuration
    ]
    for route_config in route_configs:
        for vhost in route_config["virtual_hosts"]:
            if vhost["name"] == virtual_host:
                safe_response = jsonable_encoder(vhost)
                try:
                    return json_response_class(content=safe_response)
                except TypeError:
                    return JSONResponse(content=safe_response)
        break
    return JSONResponse(content={})



from enum import StrEnum

from sovereign.utils.crypto.suites.aes_gcm_cipher import AESGCMCipher
from sovereign.utils.crypto.suites.base_cipher import CipherSuite
from sovereign.utils.crypto.suites.disabled_cipher import DisabledCipher
from sovereign.utils.crypto.suites.fernet_cipher import FernetCipher


class EncryptionType(StrEnum):
    FERNET = "fernet"
    AESGCM = "aesgcm"
    DISABLED = "disabled"


__all__ = [
    "AESGCMCipher",
    "CipherSuite",
    "DisabledCipher",
    "FernetCipher",
    "EncryptionType",
]



from enum import StrEnum
from typing import Any, Mapping, MutableMapping, Tuple, Union

EventDict = MutableMapping[str, Any]
ProcessedMessage = Union[Mapping[str, Any], str, bytes, Tuple[Any, ...]]


class LoggingType(StrEnum):
    ACCESS = "access"
    APPLICATION = "application"



from typing import Type
from importlib.util import find_spec
from fastapi.responses import JSONResponse


json_response_class: Type[JSONResponse] = JSONResponse
if find_spec("orjson"):
    from fastapi.responses import ORJSONResponse

    json_response_class = ORJSONResponse

elif find_spec("ujson"):
    from fastapi.responses import UJSONResponse

    json_response_class = UJSONResponse


__all__ = ["json_response_class"]



from functools import cached_property
from typing import Any, Dict

import structlog
from structlog.stdlib import BoundLogger

from sovereign.logging.base_logger import BaseLogger
from sovereign.logging.types import EventDict, LoggingType, ProcessedMessage
from sovereign.schemas import SovereignConfigv2


class AccessLogger(BaseLogger):
    def __init__(self, root_logger: BoundLogger, config: SovereignConfigv2):
        self._access_logs_enabled = config.logging.access_logs.enabled
        self._ignore_empty = config.logging.access_logs.ignore_empty_fields
        self._user_log_fmt = config.logging.access_logs.log_fmt

        self.logger: BoundLogger = structlog.wrap_logger(
            root_logger,
            wrapper_class=structlog.BoundLogger,
            processors=[
                self.is_enabled_processor,
                self.merge_in_threadlocal,
                self.format_access_log_fields,
            ],
            type=LoggingType.ACCESS,
        )

    @cached_property
    def is_enabled(self) -> bool:
        return self._access_logs_enabled

    @cached_property
    def _default_log_fmt(self) -> Dict[str, str]:
        return {
            "type": "{type}",
            "event": "{event}",
            "env": "{ENVIRONMENT}",
            "site": "{HOST}",
            "method": "{METHOD}",
            "uri_path": "{PATH}",
            "uri_query": "{QUERY}",
            "src_ip": "{SOURCE_IP}",
            "src_port": "{SOURCE_PORT}",
            "pid": "{PID}",
            "user_agent": "{USER_AGENT}",
            "bytes_in": "{BYTES_RX}",
            "bytes_out": "{BYTES_TX}",
            "status": "{STATUS_CODE}",
            "duration": "{DURATION}",
            "request_id": "{REQUEST_ID}",
            "resource_version": "{XDS_CLIENT_VERSION} -> {XDS_SERVER_VERSION}",
            "resource_names": "{XDS_RESOURCES}",
            "envoy_ver": "{XDS_ENVOY_VERSION}",
            "traceback": "{TRACEBACK}",
            "error": "{ERROR}",
            "detail": "{ERROR_DETAIL}",
        }

    def format_access_log_fields(
        self, logger: BoundLogger, method_name: str, event_dict: EventDict
    ) -> ProcessedMessage:
        formatted_dict: Dict[str, Any] = dict()
        for k, v in self.get_configured_log_format.items():
            try:
                value: str = v.format(**event_dict)
            except KeyError:
                value = "-"
            if value in (None, "-") and self._ignore_empty:
                continue
            formatted_dict[k] = value
        return formatted_dict
